# SCRIPT FOR SENDING PUSHOVER NOTIFICATIONS WITH A SNAPSHOT (OR GRABBED) IMAGE ATTACHMENT
#
# Version: 1.0
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V1.0    - 2023-10-14 - initial release
#
# USAGE:      {this script name}.ps1 Msg [Priority AttachFile Quality Width, TargetKB Quiet]
# ARGUMENTS:  -Msg        = [REQUIRED] [string] Pushover notification message - may contain HTML
#             -Title      = [OPTIONAL] (string) Pushover notification title (default is 'from $ScriptName')
#             -CamName    = [OPTIONAL] (string) Blue Iris camera shortname, or groupname
#             -Grab       = [OPTIONAL] (switch) instead of creating snapshot, grabs current image from camera
#             -Mainstream = [OPTIONAL] (switch) (use w/ -Grab) grabs mainstream image, if camera is dual-stream  (substream by default)
#             -Priority   = [OPTIONAL] (integer, or '*') -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal (default), 1=high-priority, 2=emergency
#             -Quality    = [OPTIONAL] (integer, or '*') target image quality 1-100; units %
#             -Width      = [OPTIONAL] (integer, or '*') target image size, smallest axis; units pixels
#             -TargetKB   = [OPTIONAL] (integer, or '*') target filesize in kilobytes
#             -Quiet      = [OPTIONAL] (switch) 'Y' or 'N') 'Y' disables background messaging
#             -Demo       = [OPTIONAL] (switch) 'Y' or 'N') 'Y' disables sending notifications, and posting to the Blue logfile
#             -NoBiLog    = [OPTIONAL] (switch) disables logging to the blue Iris logfile (enabled by default)
#
# EXAMPLES:
# 	Send notification without a snapshot attachment
#   	pushover_snapshot.ps1 -Msg 'test msg'
#   	pushover_snapshot.ps1 -Msg 'test msg' -Title "test title"
# 	Send notification without an snapshot with custom title and priority
#   	pushover_snapshot.ps1 -Msg 'test msg' -Title "test title" -Priority -1
# 	Send notification with default priority and with a camera snapshot attachment as-s (not resized)
#   	pushover_snapshot.ps1 -Msg 'test msg' * -CamName "DW1" -Quality 100 -Width 0
#   	pushover_snapshot.ps1 -Msg 'test msg' -Title "test title" -CamName "DW1" -Quality 100 -Width 0
# 	Send notification with default priority and with a group snapshot attachment as-s (not resized)
#   	pushover_snapshot.ps1 -Msg 'test msg' * -CamName "Outdoor" -Quality 100 -Width 0
#   	pushover_snapshot.ps1 -Msg 'test msg' -Title "test title" -CamName "Outdoor" -Quality 100 -Width 0
# 	Change only img_quality, use default img_width
#   	pushover_snapshot.ps1 -Msg 'test msg' * -CamName "Outdoor" -Quality 70
# 	Change only img_width, use default img_quality
#		pushover_snapshot.ps1 -Msg 'test msg' -CamName "Outdoor" -Width 500
# 	Change both img_quality and img_width
#		pushover_snapshot.ps1 -Msg 'test msg' * -CamName "Outdoor" -Quality 70 -Width 500
# 	Target attachment size of 500 KB, using default img_quality & img_width
#		pushover_snapshot.ps1 'test msg' -CamName "Outdoor" -TargetKB 500
# 	Target attachment size of 500 KB, using specified img_quality
#		pushover_snapshot.ps1 -Msg 'test msg' -CamName "Outdoor" -Quality 70 -TargetKB 500

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[string]$Msg,         # [REQUIRED]
	[string]$CamName,     # [REQUIRED]
	[string]$Title,     
	[Switch]$Grab,  
	[Switch]$MainStream,  
	[int]$Priority,       
	[string]$Quality,     
	[string]$Width,       
	[string]$TargetKB,    
	[switch]$Quiet,       
	[switch]$Demo,        
	[switch]$Help,
	[switch]$NoBiLog,
	[switch]$DevMode
)

# For debugging
#Write-Host "Msg = $Msg"
#Write-Host "CamName = $CamName"
#Write-Host "Title = $Title"
#Write-Host "MainStream = $MainStream"
#Write-Host "Priority = $Priority"
#Write-Host "Quality = $Quality"
#Write-Host "Width = $Width"
#Write-Host "TargetKB = $TargetKB"
#Write-Host "Title = $Title"
#Write-Host "Quiet = $Quiet"
#Write-Host "Demo = $Demo"
#Write-Host "NoBiLog = $NoBiLog"
#Write-Host "DevMode = $DevMode"

#Set-StrictMode  -Version 3
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

# GATHER SCRIPT INFO 
$ScriptName = $MyInvocation.MyCommand.Name
$ScriptBaseName = $ScriptName.Replace('.ps1', '')
$ScriptPath = Split-Path $MyInvocation.MyCommand.Path

# ================================================
# GET USER SETTINGS...

# Import external 'user settings' script (expected to be in same folder as this script)
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	$err_msg = "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
	return $err_msg
}

# END GET USER SETTINGS
# ===============================================

$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$app_path = "C:\temp\pushover_snapshot\"
} else {
	if (!$app_path.EndsWith('\')) {
		$app_path = $app_path + '\'
	}
}
$variable_test = Get-Variable -Name bi_new_path -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$err_mg = "ERR/User settings file lacks setting for Blue Iris New folder"
	return $err_msg
} else {
	if (!$bi_new_path.EndsWith('\')) {
		$bi_new_path = $bi_new_path + '\'
	}
	if (!(Test-Path -LiteralPath "$bi_new_path")) {
		$err_mg = "ERR/Cannot find Blue Iris New folder"
		return $err_msg
	}
}
$variable_test = Get-Variable -Name bg_msgs_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$bg_msgs_flg = $true
}
$variable_test = Get-Variable -Name show_method_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$show_method_flg = $true
}
$variable_test = Get-Variable -Name snap_delay_ms -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$snap_delay_ms = 500
}
$variable_test = Get-Variable -Name snap_memo -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$snap_memo = $ScriptBaseName
}
$variable_test = Get-Variable -Name po_title -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$po_title = "from $ScriptName"
}
$variable_test = Get-Variable -Name po_priority -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$po_priority = 0
}
$variable_test = Get-Variable -Name targetkb_method -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$targetkb_method = 'WxH'
}
if (@('WxH','Quality','Hybrid') -notcontains $targetkb_method) {
	$targetkb_method = 'WxH'
}
$variable_test = Get-Variable -Name targetkb_wxh_step -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$targetkb_wxh_step = 25
}
$variable_test = Get-Variable -Name targetkb_quality_step -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$targetkb_quality_step = 20
}
$variable_test = Get-Variable -Name img_quality -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$img_quality = 90
}
$variable_test = Get-Variable -Name img_width -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$img_width = 500
}
$variable_test = Get-Variable -Name img_kilobytes -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$img_kilobytes = 0
}
$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$post_to_BI_logfile_flg = $false
}
$variable_test = Get-Variable -Name post_all_events -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$post_all_events = $false
}

# INITIALIZE VARS
$PsVersion = $PSVersionTable.PSVersion.Major
# NOTE: Scripts running from Windows Task Scheduler have a parent process named 'svchost'.
#       https://stackoverflow.com/a/65894924
$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
# NOTE: Scripts called by another script have a 'PSCommandPath'.
#       Method 'PSCommandPath' returns null if script is executed via the terminal
#       https://stackoverflow.com/a/23370618/13499984
$CallingScript = $MyInvocation.PSCommandPath

# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
if ($ParentProcess -eq 'BlueIris') {
	$fromConsole = $false
} else {
	$fromConsole = ($CallingScript -eq $null)
}

# For debugging
#msg.exe * /TIME:10 "entered script '$ScriptName'"
#msg.exe * /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
#msg.exe * /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
#msg.exe * /TIME:10 "fromConsole ($ScriptName) = $fromConsole"

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if (!$fromConsole) {
	if ($bg_msgs_flg ) {
		$ExecPolicy = Get-ExecutionPolicy
		$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
		if ($env:PSExecutionPolicyPreference) {
			$ExecPolicyPref = $env:PSExecutionPolicyPreference
			$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
		}
		$msg_sec = 10
		$bg_msg = "Script '$ScriptName' is running in the background" + 
		"`n$exec_policy_str" + 
		"`nPsVersion = $PsVersion" + 
		"`nParentProcess = $ParentProcess" + 
		"`nalert = $AttachFile" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		msg.exe * /TIME:$msg_sec $bg_msg
	}
}

# For testing CLI mode
#$fromConsole = $false

# ===============================================

function BIJsonCall() {
# PURPOSE:   To get a Blue Iris JSON interface response from the Blue Iris web server.
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -JsonCmd        = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
#            -ForceFlg       = [OPTIONAL] (boolean) forces new login (gets a new authenticated session id)
#            -QuietFlg       = [OPTIONAL] (boolean) $true suppresses verbose output to the console
#            -FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS:   (object) PowerShell PSCustomObject
# REFS:      1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
#            2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
#            3) https://ipcamtalk.com/threads/json-interface.21491/
#            4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
#
# Created by jaydeel at ipcamtalk.com

	param (
		[hashtable]$JsonCmd,
		[boolean]$ForceFlg,
		[boolean]$QuietFlg,
		[boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name() ..."	
	#Write-Host "`Entered function $fn_name()`n------------------------"
	#Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"	

	# System vars
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	$PsVersion = $PSVersionTable.PSVersion.Major
	# Get the current server domain (hostname) and username
	# Used for background messaging via msg.exe
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$server_name = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	function encryptMD5($str) {
	# PURPOSE: Get MD5 hash for a string.
	# ARGS:    $str = (string) string to be encrypted
	# RETURNS: (string) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
	# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
	# NOTES:   (from REF)
	#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
	#	However, PowerShell offers a cmdlet to generate MD5 hash for a file: Get-FileHash. 
	# 	The cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
	# UTILITY: https://www.md5hashgenerator.com/
	# ... Example: 'MD5string' returns '0753a0e20c9027ccab70f4291a21aa08'

		$stringAsStream = [System.IO.MemoryStream]::new()
		$writer = [System.IO.StreamWriter]::new($stringAsStream)
		$writer.write($str)
		$writer.Flush()
		$stringAsStream.Position = 0
		$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
		
		# For debugging
		#$str = $hash_obj | Out-String
		#Write-Host "hash_obj = $str"

		return $hash_obj.Hash
	}

	# HANDLE SAVED SESSION IDS
	# ... Perform existence test for the *.txt file used to save session ids
	#     Create its path if it does not exist
	#     NOTE: if a path is not specified, the default Windows location is 'C:\Users\<username>"
	$path = "c:\temp\$ScriptBaseName"
	if (!(Test-Path -LiteralPath $path)) {
		New-Item -ItemType Directory -Force -Path $path
	}
	# ... NOTE: we use a different filename when running the script via different processes;
	#     this avoids collisions with scripts running via:
	#     'scvhost' (Task Scheduler), 'WindowsTerminal', 'explorer' (Powershell ISE)
	$base_name = "$path\bi_session_id"
	$session_filespec = $base_name + "_" + $ParentProcess + ".txt"
	$found_saved_session_file_flg = (Test-Path -LiteralPath "$session_filespec")

	# PARSE COMMANDLINE ARGS
	if ($JsonCmd -eq $null) {
		$err_msg = "ERR/expecting an argument: JsonCmd (PowerShell hashtable)`n"
		return $err_msg
	} else {
		# For debugging
		#Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
	}
	# ... Save original JsonCmd (in case we need to make a recursive call)
	$json_cmd_orig = $JsonCmd
	# ... Override reuse of saved session_id
	if ($ForceFlg) {
		$found_saved_session_file_flg = $false
	}

	# ========================
	# PERFORM API CALLS

	# Make the URL
	$uri = "http://" + $script:bi_webserver + "/json";

	# For debugging
	#Write-Host "$fn_name() uri = $uri"

	# Flag 'logout' JSON cmd requests
	# ... NOTE: All JSON 'logout' cmds should be performed by the calling script!
	if ($JsonCmd.cmd -eq 'logout') {
		$logging_out_flg = $true
	} else {
		$logging_out_flg = $false
	}
	
	# For debugging
	if ($logging_out_flg) {
		#Write-Host "$fn_name() detected JSON cmd to log out of BI API session"
	} else {
		#Write-Host "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
		#msg.exe * /TIME:5 "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
	}

	# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
	$dev_uuid = $null  # Default
	$json_cmd_keys = $JsonCmd.Keys
	if ($json_cmd_keys -contains 'uuid') {
		if ($JsonCmd.uuid -ne $null) {
			$dev_uuid = $JsonCmd.uuid
		}
	}

	# For debugging
	if ($dev_uuid) {
		#Write-Host "$fn_name() dev_uuid = $dev_uuid"
	}

	# EXECUTE API CALL
	
	if (!$found_saved_session_file_flg) {
		
		# ================================================================
		# 1.1 PRELIMINARY LOGIN
		
		# Create, then encode the JSON 'login' command
		$parameters = @{
			cmd = 'login'
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$jsonResponseObj"
		
		# Get the session id
		$session = $jsonResponseObj.session
		
		# ================================================================
		# 1.2 AUTHENTICATE THE LOGIN

		# Create MD5 hash-encoded credentials string
		$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
		$response_hash = encryptMD5 $response_str

		# Create, then encode a modified JSON 'login' command
		if ($dev_uuid) {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
				uuid = $dev_uuid
			}
		} else {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
			}
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-Host "$fn_name() (1.2) authorizing json login... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) authorizing json login ... `n$json_params"
		
		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#Write-Host ("$fn_name() (1.2) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) jsonResponse... `n$jsonResponseObj"
		
		# Handle failure
		if ($jsonResponseObj.result -eq 'fail') {
			$err_msg = "ERR/unable to authenticate Blue Iris JSON interface`n"
			return $err_msg
		}
		
		# Get the authenticated session id
		$auth_session_id = $jsonResponseObj.session

		# For debugging
		#msg.exe $server_username /TIME:5 "$fn_name() auth_session_id... $auth_session_id"
	
		# Save the authenticated session id
		# NOTE: the session id is always saved, whether used or not
		$auth_session_id | Set-Content "$session_filespec" -NoNewline

	} else {
	
		# Retrieve the saved session id
		# KEY POINT: 
		#   Force typing the output of 'Get-Content' to [string] was essential when running script in
		#   in Powershell ISE. Otherwise the output was similar to an object.
		#   *** But in actuality it is more complicated ***
		#   Per https://stackoverflow.com/a/55881978/13499984
		#      When 'Get-Content' is assigned to a variable, its output is an array of .NET [string] objects, 
		#      each representing a line from the input file. However, Get-Content also decorates these string 
		#      objects with additional properties that provide helpful metadata about where the strings came from.
		$saved_session_id = [string](Get-Content -Path "$session_filespec")

		$auth_session_id = $saved_session_id

		# For debugging
		if (!$logging_out_flg) {
			#Write-Host ("$fn_name() saved_session_id type = " + $saved_session_id.GetType())
			#Write-Host "$fn_name() saved_session_id... $saved_session_id"
			#msg.exe * /TIME:5 "$fn_name() saved_session_id... $saved_session_id"
		}
		
	}

	# ================================================================
	# 1.3 EXECUTE THE REQUESTED JSON COMMAND

	if ($dev_uuid -eq $null) {
		
		# Modify, then encode the requested JSON login command
		$JsonCmd.session = $auth_session_id
		$json_params = $JsonCmd | ConvertTo-Json

		$max_retries = 3
		$cntr = 1
		while ($cntr -le $max_retries) {

			# For debugging
			#Write-Host ("$fn_name() (1.3) loop #" + $cntr + "...")

			# Execute the JSON request
			$params = @{
				Uri                = $uri
				#Headers            = @{"Content-Type"  = "application/json"}
				Method             = 'POST'
				#Body               = $JsonCmd | ConvertTo-Json
				Body               = $json_params
				ContentType        = 'application/json'
			}
			
			$jsonResponseObj = Invoke-RestMethod @params

			# Alternate syntaxes (pick one)
			#$jsonResponseObj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
			#$jsonResponseObj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

			# For debugging
			#if ($JsonCmd.cmd -match 'update') {
				#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "'")
				#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "' ...`n" + $json_params)
				#Write-Host ("$fn_name() jsonResponseObj = `n" + ($jsonResponseObj | ConvertTo-Json))
			#}

			if ($jsonResponseObj.result -eq 'fail') {
				
				# For debugging
				#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response.result = fail")
				#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response = `n" + ($jsonResponseObj | ConvertTo-Json))

				$data = $jsonResponseObj.data

				# For debugging
				#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data = " + ($data | ConvertTo-Json))

				# Try to determine reason
				if ($data) {
					if ($data.reason) {
						$reason = $data.reason
					
						# For debugging
						#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data.reason = $reason")
						
						if ($reason -match 'Access denied') {
							Write-Warning ("JsonCmd '" + $JsonCmd.cmd + "' requires an Admin account")
							$err_msg =" ERR/JsonCmd '" + $JsonCmd.cmd + "' requires an Admin account"
							return $err_msg
						} else {
							Write-Warning ("JsonCmd '" + $JsonCmd.cmd + "' failed with reason '" + $reason + "'")
							$err_msg =" ERR/JsonCmd '" + $JsonCmd.cmd + "' failed with reason '" + $reason + "'"
							return $err_msg
						}
					} else {
						Write-Warning ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response is 'fail' without a reason - trying again - retry $cntr (of $max_retries max)")
					}
				}
					
				# For debugging
				#Write-Host "$fn_name() (1.3) performing RECURSIVE JSON call..."
			
				# Recursive call
				$json_response_obj = `
					BIJsonCall `
					-JsonCmd $json_cmd_orig `
					-ForceFlg $true `
					-QuietFlg $QuietFlg `
					-FromConsoleFlg $FromConsoleFlg
					
				# For debugging
				#Write-Host ("$fn_name() (1.3) JsonCmd '" + $json_cmd_orig.cmd + "' RECURSIVE response.result = " + $json_response_obj.result)
				#Write-Host ("$fn_name() (1.3) JsonCmd '" + $json_cmd_orig.cmd + "' RECURSIVE response = `n" + ($json_response_obj | ConvertTo-Json))
		
				# Import the new session id (due to -ForceFlg)
				$saved_session_id = [string](Get-Content -Path "$session_filespec")
				$auth_session_id = $saved_session_id

				# Modify, then encode the requested JSON login command
				$JsonCmd.session = $auth_session_id
				$json_params = $JsonCmd | ConvertTo-Json
				
				if ($json_response_obj.result -eq 'fail') {
					Write-Host "RECURSIVE RESULT 'FAIL', TRYING AGAIN..."
					$cntr++
					Start-Sleep -Milliseconds 100
					continue
				}
				
			} else {

				# Test for bad response (null, empty)
				if ($PsVersion -ge 7) {
					# Avoids the following WARNING...
					# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
					$jro = $jsonResponseObj | ConvertTo-Json -Depth 16
				} else {
					$jro = $jsonResponseObj | ConvertTo-Json
				}
				$bad_response_flg = (($jro -eq $null) -or $($jro -eq "null") -or $($jro -eq "") -or ($jro -eq "{}"))
				
				if (!$bad_response_flg) {
					# Response is good. All done.

					# For debugging
					#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response.result = success")
					#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response  = $jro")

					break
					
				} else {

					# For debugging
					#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' bad response  = $jro")

					# Automatically retry for a limited set of times
					if ($cntr -eq $max_retries) {
						if ($FromConsoleFlg -and !$QuietFlg) {
							$err_msg = "$fn_name() (1.3) quit retries after $max_retries iterations"
							return $err_msg
						}
					} else {
						if ($FromConsoleFlg -and !$QuietFlg) {
							$err_msg = "$fn_name() (1.3) BI API response empty or null - retry $cntr (of $max_retries max)"
							return $err_msg
						}
						$cntr++
						Start-Sleep -Milliseconds 100
						continue
					}
				}
			}
		}
	}

	# For debugging
	if ($jsonResponseObj) {
		#Write-Host ("$fn_name() (1.3) jsonResponse type = " + $jsonResponseObj.GetType())
		#Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $jsonResponseObj"
	} 

	# Retain for error testing
	#$jsonResponseObj = $null
	#$jsonResponseObj = ''

	if ($jsonResponseObj -eq $null) {
		$err_msg ="ERR/null response from Blue Iris JSON interface"
		return $err_msg
	} 
	if ($jsonResponseObj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris JSON interface"
		return $err_msg
	}
	if ($jsonResponseObj.result -ne $null) {

		# For debugging
		#Write-Host ("$fn_name() (1.3) jsonResponseObj.result... " + ($jsonResponseObj.result))
		
		if ($jsonResponseObj.result -eq 'fail') {
			
			# Show the contents of the $jsonResponseObj 'data' key

			$json_data_obj = $jsonResponseObj.data
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) jsonResponseObj.data... " + ($jsonResponseObj.data))

			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jdo_str = $json_data_obj | ConvertTo-Json -Depth 16
			} else {
				$jdo_str = $json_data_obj | ConvertTo-Json
			}
			
			# 'Minify' the json string.
			if ($jdo_str) {
				if (!($jdo_str -match 'null')) {
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is 'result:fail'"
					return $err_msg
				} else {					
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is null"
					return $err_msg
				}
			}
		}
	}

	return $jsonResponseObj

}

function PostCustomBiLogEntry() {
# PURPOSE: Add custom message to the Blue Iris logfile
# ARGS:    $MsgStr        = (string) may contain HTML
#          $Level         = (integer) 0|1|2 (0:info, 1:warn, 2:error)
#          $SilentFlg     = (boolean) $true prevents background messaging
# RETURNS: nothing

    param(
		[string]$MsgStr,      # may contain HTML
		[string]$Level,       # full filespec of resized alert image
		[boolean]$SilentFlg   # 
	)

	# For debugging
	#Write-Host "SilentFlg = $SilentFlg"
	
	# Remove/replace incompatible substrings from $MsgStr
	$MsgStr = $MsgStr -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"
	
	$bi_address = $script:bi_webserver
	if ($script:post_w_creds_flg) {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level&user=$script:bi_user_name&pw=$script:bi_user_pw"
	} else {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level"
	}
	$result = Invoke-WebRequest -Uri $uri
	
	# For debugging
	#Write-Host "uri = $uri"
	#Write-Host "result:`n$result"	

	# HANDLE NOTIFICATIONS
	if (!$SilentFlg) {
		msg.exe * /TIME:10 "Custom BI logfile message posted...`nlevel = $Level`n'$MsgStr'"
	}
	if ($script:fromConsole) {
		Write-Host ("`nCustom BI logfile message posted:`n'" + ($MsgStr.Replace("%20"," ")) + "'`n")
	}
}

function TakeSnapshot() {
# PURPOSE:  Take a Blue Iris camera snapshot using HTTP interface command "/image/{cam-shortname}"
# ARGS:     see param lines below
# RETURNS:  'success', OR error message
	
    param(
		[string]$CamName    # Blue Iris camera shortname, or groupname (case-sensitive)
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"	
	#Write-Host "$fn_name() CamName = '$CamName'"

	# ========================
	# PERFORM API CALL

	# Make the URL

	# NOTE: this commands does not work with groupnames
	#$HttpCmd = "/admin?camera=" + $CamName + "&snapshot"

	$HttpCmd = "/admin?camera=" + $CamName + "&snapshot"
	if ($use_bi_creds) {
		$uri = "http://" + $bi_webserver + $HttpCmd + "&user=" + $bi_user_name + "&pw=" + $bi_user_pw
	} else {
		$uri = "http://" + $bi_webserver + $HttpCmd	
	}

	# For debugging
	#Write-Host "uri = '$uri'"

	# ================================================================
	# EXECUTE THE REQUESTED HTTP COMMAND 

	$web_response_obj = $null  # Assume failure

	try {

		$web_response_obj = Invoke-WebRequest -Uri $uri

	} 
	catch {

		# For debugging
		#Write-Host $error
		#Write-Host $_.ScriptStackTrace
		#Write-Host $PSitem

		Write-Host $_.Exception.Message -ForegroundColor Red
		$err_msg = "ERR/unexpected HTTP Interface command response"
		return $err_msg
	}

	# For debugging
	# REF: https://4sysops.com/archives/powershell-invoke-webrequest-parse-and-scrape-a-web-page/
	if ($web_response_obj) {
		<# Write-Host "Web Response..."
		Write-Host ("... type = " + $web_response_obj.GetType())
		Write-Host ("... StatusCode = " + $web_response_obj.StatusCode)
		Write-Host ("... StatusDescription = " + $web_response_obj.StatusDescription)
		#Write-Host ("... Headers = " + $web_response_obj.Headers)
		Write-Host ("... Content-Type = " + $web_response_obj.Headers['Content-Type'])
		Write-Host ("... Content-Length = " + $web_response_obj.Headers['Content-Length'])
		#Write-Host ("... Content = " + $web_response_obj.Content) #>
	} 

	# @@@ DEV - Retain for error testing
	#$web_response_obj = $null
	#$web_response_obj = ''

	if ($web_response_obj -eq $null) {
		$err_msg = "ERR/null response from Blue Iris HTTP interface"
		return $err_msg
	} 
	if ($web_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris HTTP interface"
		return $err_msg
	}
	If ($web_response_obj -match "<!DOCTYPE html>") {

		# For debugging
		#$HTML = New-Object -Com "HTMLFile"
		#$HTML.write([ref]$web_response_obj)
		#$TextOnly = $Html.body.innerText
		#Write-Host $TextOnly -ForegroundColor Cyan

		$err_msg = "ERR/unexpected response"
		return $err_msg
	}
	
	return "success"
	
}

function GrabStreamImage() {
# PURPOSE:  Get Blue Iris camera/group 'snapshot' using HTTP interface command "/image/{cam-shortname}"
# ARGS:     see param lines below
# RETURNS:  'success', OR error message
	
    param(
		[string]$CamName,         # Blue Iris camera shortname, or groupname (case-sensitive)
		[boolean]$MainStreamFlg,  # $true grabs mainstream image (substream by default)
		[string]$DestPath         # full filespec for 'snapshot' image
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"	
	#Write-Host "$fn_name() CamName = '$CamName'"
	#Write-Host "$fn_name() MainStreamFlg = '$MainStreamFlg'"
	#Write-Host "$fn_name() DestPath = '$DestPath'"

	# ========================
	# PERFORM API CALL

	# Make the URL

	# NOTE: this commands does not work with groupnames
	#$HttpCmd = "/admin?camera=" + $CamName + "&snapshot"

	if ($MainStreamFlg) {
		$HttpCmd = "/image/" + $CamName + "&decode=1"
	} else {
		$HttpCmd = "/image/" + $CamName
	}
	<# if ($use_bi_creds) {
		$uri = "http://" + $bi_webserver + $HttpCmd + "&user=" + $bi_user_name + "&pw=" + $bi_user_pw
	} else {
		$uri = "http://" + $bi_webserver + $HttpCmd	
	} #>
	
	$uri = "http://" + $bi_webserver + $HttpCmd	

	# For debugging
	#Write-Host "uri = '$uri'"

	# ================================================================
	# EXECUTE THE REQUESTED HTTP COMMAND 

	$web_response_obj = $null  # Assume failure

	try {

		Invoke-WebRequest -Uri $uri -OutFile $DestPath

		# Repeat to capture response for debugging
		$web_response_obj = Invoke-WebRequest -Uri $uri

	} 
	catch {

		# For debugging
		#Write-Host $error
		#Write-Host $_.ScriptStackTrace
		#Write-Host $PSitem

		Write-Host $_.Exception.Message -ForegroundColor Red
		$err_msg = "ERR/unexpected HTTP Interface command response"
		return $err_msg
	}

	# For debugging
	# REF: https://4sysops.com/archives/powershell-invoke-webrequest-parse-and-scrape-a-web-page/
	if ($web_response_obj) {
		<# Write-Host "Web Response..."
		Write-Host ("... type = " + $web_response_obj.GetType())
		Write-Host ("... StatusCode = " + $web_response_obj.StatusCode)
		Write-Host ("... StatusDescription = " + $web_response_obj.StatusDescription)
		#Write-Host ("... Headers = " + $web_response_obj.Headers)
		Write-Host ("... Content-Type = " + $web_response_obj.Headers['Content-Type'])
		Write-Host ("... Content-Length = " + $web_response_obj.Headers['Content-Length'])
		#Write-Host ("... Content = " + $web_response_obj.Content) #>
	} 

	# @@@ DEV - Retain for error testing
	#$web_response_obj = $null
	#$web_response_obj = ''

	if ($web_response_obj -eq $null) {
		$err_msg = "ERR/null response from Blue Iris HTTP interface"
		return $err_msg
	} 
	if ($web_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris HTTP interface"
		return $err_msg
	}
	If ($web_response_obj -match "<!DOCTYPE html>") {

		# For debugging
		#$HTML = New-Object -Com "HTMLFile"
		#$HTML.write([ref]$web_response_obj)
		#$TextOnly = $Html.body.innerText
		#Write-Host $TextOnly -ForegroundColor Cyan

		$err_msg = "ERR/unexpected response"
		return $err_msg
	}
	
	return "success"
	
}

function GetLatestSnapshotFile() {
# PURPOSE:  Get Blue Iris camera snapshot using HTTP interface command "/image/{cam-shortname}"
# ARGS:     see param lines below
# RETURNS:  snapshot file full filespec
	
    param(
		[string]$CamName,    # Blue Iris camera shortname, or groupname (case-sensitive)
		[string]$SourcePath  # Path for Blue Iris snapshot JPGS ('New' folder)
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"	
	#Write-Host "$fn_name() CamName = '$CamName'"
	#Write-Host "$fn_name() SourcePath = '$SourcePath'"

	# ========================
	# PERFORM API CALL

	# Make the URL

	# NOTE: this commands does not work with groupnames
	#$HttpCmd = "/admin?camera=" + $CamName + "&snapshot"

	$HttpCmd = "/admin?camera=" + $CamName + "x&snapshot"
	<# if ($use_bi_creds) {
		$uri = "http://" + $bi_webserver + $HttpCmd + "&user=" + $bi_user_name + "&pw=" + $bi_user_pw
	} else {
		$uri = "http://" + $bi_webserver + $HttpCmd	
	} #>
	
	$uri = "http://" + $bi_webserver + $HttpCmd	

	# For debugging
	#Write-Host "uri = '$uri'"

	# ================================================================
	# EXECUTE THE REQUESTED HTTP COMMAND 

	$web_response_obj = $null  # Assume failure

	try {

		$web_response_obj = Invoke-WebRequest -Uri $uri

	} 
	catch {

		# For debugging
		#Write-Host $error
		#Write-Host $_.ScriptStackTrace
		#Write-Host $PSitem

		Write-Host $_.Exception.Message -ForegroundColor Red
		$err_msg = "ERR/unexpected HTTP Interface command response"
		return $err_msg
	}

	# For debugging
	# REF: https://4sysops.com/archives/powershell-invoke-webrequest-parse-and-scrape-a-web-page/
	if ($web_response_obj) {
		<# Write-Host "Web Response..."
		Write-Host ("... type = " + $web_response_obj.GetType())
		Write-Host ("... StatusCode = " + $web_response_obj.StatusCode)
		Write-Host ("... StatusDescription = " + $web_response_obj.StatusDescription)
		#Write-Host ("... Headers = " + $web_response_obj.Headers)
		Write-Host ("... Content-Type = " + $web_response_obj.Headers['Content-Type'])
		Write-Host ("... Content-Length = " + $web_response_obj.Headers['Content-Length'])
		#Write-Host ("... Content = " + $web_response_obj.Content) #>
	} 

	# @@@ DEV - Retain for error testing
	#$web_response_obj = $null
	#$web_response_obj = ''

	if ($web_response_obj -eq $null) {
		$err_msg = "ERR/null response from Blue Iris HTTP interface"
		return $err_msg
	} 
	if ($web_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris HTTP interface"
		return $err_msg
	}
	If ($web_response_obj -match "<!DOCTYPE html>") {

		# For debugging
		#$HTML = New-Object -Com "HTMLFile"
		#$HTML.write([ref]$web_response_obj)
		#$TextOnly = $Html.body.innerText
		#Write-Host $TextOnly -ForegroundColor Cyan

		$err_msg = "ERR/unexpected response"
		return $err_msg
	}
	
	return "success"
	
}

function BIJsonCliplist() {
# PURPOSE: Get Blue Iris web server JSON 'cliplist' command response for the specified camera
#          Intended to be called by other PowerShell scripts.
# ARGS:    $CamName   = [REQUIRED] camera shortname
#          $Props     = [OPTIONAL] (list) like 'file,date', comma delimited
#                                   choices: camera, path, file, offset, date, color, flags, res, msec, filesize, filetype
#          $DbView    = [OPTIONAL] (string) choose one of the following:
#                                  all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
#                                  flagged, export, archive, people, vehicles, confirmed, or canceled
#                                  (if unspecified, the default is 'alerts')
#          $LatestN   = [OPTIONAL] (integer) returns the N most recent clips only
#          $FileType  = [OPTIONAL] (string) like 'bvr New' or 'jpg New'
#          $ClipPath  = [OPTIONAL] (string) like '@123456789'
#          $Beg       = [OPTIONAL] (string) datetime like 'YYYYMMDDHHMM' ... TO BE IMPLEMENTED
#          $End       = [OPTIONAL] (string) datetime like 'YYYYMMDDHHMM' ... TO BE IMPLEMENTED
#          $QuietFlg  = [OPTIONAL] (boolean) true suppresses verbose informational output to the terminal
# RETURNS: (object) PowerShell hashtable, @('cam':camName,'alertProp':value}
#
# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
# "returns a list of cameras on the system ordered by group. Cameras not belonging to any
# group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
# the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
# describing a camera or a camera group."
#
# Version: X
# Created by jaydeel at ipcamtalk.com

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[string]$CamName,          # [REQUIRED]
		[string]$Props,            # [OPTIONAL] list
		[string]$DbView,           # [OPTIONAL] default is 'all'
		[int]$LatestN,             # [OPTIONAL]
		[string]$ClipPath,         # [OPTIONAL] 
		[string]$FileType,         # [OPTIONAL] 
		[int]$Beg,                 # [OPTIONAL] 
		[int]$End,                 # [OPTIONAL] 
		[boolean]$FromConsoleFlg,  # [OPTIONAL]
		[boolean]$QuietFlg         # [OPTIONAL]
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() CamName = $CamName"
	#Write-Host "$fn_name() Props = $Props"
	#Write-Host "$fn_name() ClipPath = $ClipPath"
	#Write-Host "$fn_name() DbView = $DbView"
	#Write-Host "$fn_name() LatestN = $LatestN"
	#Write-Host "$fn_name() ClipPath = $ClipPath"
	#Write-Host "$fn_name() FilteType = $FilteType"
	#Write-Host "$fn_name() Beg = $Beg"
	#Write-Host "$fn_name() End = $End"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"

	# For debugging
	
	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$CamName) {
		$err_msg = "ERR/expecting at least 1 argument (CamName)"
		return $err_msg
	}

	# HANDLE OPTIONAL COMMANDLINE ARGS
	if ($Props) {
		$PropsArg = $Props
		$clipProps = $PropsArg -Split ','
	} else {
		$clipProps = @()
	}
	#Write-Host ("alertProps cnt = " + ($clipProps.Count))
	#Write-Host "alertProps = $clipProps"
	if (!$DbView) {
		$DbView = 'all'
	}
	$variable_test = Get-Variable -Name LatestN -Scope Local -ErrorAction SilentlyContinue
	if ($variable_test -eq $null) {
		$latest_N = [int]$LatestN
	} else {
		$latest_N = $null
	}
	$variable_test = Get-Variable -Name Beg -Scope Local -ErrorAction SilentlyContinue
	if ($variable_test -eq $null) {
		$dt_beg = $Beg
	} else {
		$dt_beg = $null
	}
	$variable_test = Get-Variable -Name End -Scope Local -ErrorAction SilentlyContinue
	if ($variable_test -eq $null) {
		$dt_beg = $End
	} else {
		$dt_beg = $null
	}

	# Initialize collectors
	$clip_props_obj = @{}
	$clip_props_obj.Add( 'camname', @() )
	$clip_props_obj.Add( 'path', @() )
	$clip_props_obj.Add( 'file', @() )
	$clip_props_obj.Add( 'offset', @() )
	$clip_props_obj.Add( 'date', @() )
	$clip_props_obj.Add( 'color', @() )
	$clip_props_obj.Add( 'flags', @() )
	$clip_props_obj.Add( 'res', @() )
	$clip_props_obj.Add( 'msec', @() )
	$clip_props_obj.Add( 'filesize', @() )
	$clip_props_obj.Add( 'filetype', @() )
	#Write-Host ("al_props = " + ($clip_props_obj | ConvertTo-Json))

	$tmp_obj = @{}
	$tmp_obj.Add( 'camname', @() )
	$tmp_obj.Add( 'path', @() )
	$tmp_obj.Add( 'file', @() )
	$tmp_obj.Add( 'offset', @() )
	$tmp_obj.Add( 'date', @() )
	$tmp_obj.Add( 'color', @() )
	$tmp_obj.Add( 'flags', @() )
	$tmp_obj.Add( 'res', @() )
	$tmp_obj.Add( 'msec', @() )
	$tmp_obj.Add( 'filesize', @() )
	$tmp_obj.Add( 'filetype', @() )

	# KEY POINT: BI's 'alertlist' json cmd returns only 1000 items; 
	#            so we must re-send the json cmd until we have retrieved all alerts.
	 
	$i = 0;
	$tmp_obj['paths'] = @('1');  # Stat with a dummy value
	while ($tmp_obj['paths'].Count -gt 0) {
		
		# For debugging
		#Write-Host "-----`niter #$i"

		# PERFORM API CALLS
		if ($i -gt 0) {
			$jsonCmd = @{
				cmd = 'cliplist'
				camera = $CamName
				view = $DbView
				enddate = $clip_props_obj['dates'] | Select-Object -Last 1
			}
		} else {
			$jsonCmd = @{
				cmd = 'cliplist'
				camera = $CamName
				view = $DbView
			}
		}
		#Write-Host ("jsonCmd = " + ($jsonCmd | ConvertTo-Json))

		# RETURNS A POWERSHELL PSCustomObject
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-QuietFlg $QuietFlg

		# For debugging
		#Write-Host ("json_response_obj type = " + $json_response_obj.GetType())
		#Write-Host ("json_response_obj = " + ($json_response_obj | ConvertTo-Json))
		#msg.exe * /TIME:10 "json_response... $json_response_obj"	

		if ($json_response_obj -match 'success') {
			$data_obj = $json_response_obj.data
			if ($data_obj) {

				# For debugging				
				#Write-Host ("End date for pass #" + ($i+1) + " (" + ($data_obj.path.Count) + " items) = " + ($data_obj.date | Select-Object -Last 1))

				$tmp_obj['camname']  = $data_obj.camera
				$tmp_obj['path']     = $data_obj.path
				$tmp_obj['file']     = $data_obj.file
				$tmp_obj['offset']   = $data_obj.offset
				$tmp_obj['date']     = $data_obj.date	  # Note: $date_ary[] is reverse-ordered.
				$tmp_obj['color']    = $data_obj.color
				$tmp_obj['flags']    = $data_obj.flags
				$tmp_obj['res']      = $data_obj.res
				$tmp_obj['msec']     = $data_obj.msec
				$tmp_obj['filesize'] = $data_obj.filesize
				$tmp_obj['filetype'] = $data_obj.filetype

				# Merge with master accumulator.
				$clip_props_obj['camname']  = $clip_props_obj['camname'] + $tmp_obj['camname']
				$clip_props_obj['path']     = $clip_props_obj['path'] + $tmp_obj['path']
				$clip_props_obj['file']     = $clip_props_obj['file'] + $tmp_obj['file']
				$clip_props_obj['offset']   = $clip_props_obj['offset'] + $tmp_obj['offset']	
				$clip_props_obj['date']     = $clip_props_obj['date'] + $tmp_obj['date']	
				$clip_props_obj['flags']    = $clip_props_obj['flags'] + $tmp_obj['flags']	
				$clip_props_obj['res']      = $clip_props_obj['res'] + $tmp_obj['res']	
				$clip_props_obj['msec']     = $clip_props_obj['msec'] + $tmp_obj['msec']	
				$clip_props_obj['filesize'] = $clip_props_obj['filesize'] + $tmp_obj['filesize']	
				$clip_props_obj['filetype'] = $clip_props_obj['filetype'] + $tmp_obj['filetype']
			} else {
				# No more clips to collect.
				break;
			}
		} else {
			break;
		}
		
		$i++
	}

	# For debugging
	#Write-Host ("$clip_props_obj type = " + ($clip_props_obj.GetType()))
	#Write-Host ("$clip_props_obj = " + ($clip_props_obj | ConvertTo-Json))
	#Write-Host ("clip_props_obj.paths = " + $clip_props_obj.paths)
	#Write-Host ("clip_props_obj keys = " + $clip_props_obj.Keys)

	# Get all clips paths and their count
	$all_clips_paths = $clip_props_obj.path
	$clips_cnt = $clip_props_obj['path'].Count

	# Get all clips filetypes 
	$all_clips_filetypes = $clip_props_obj.filetype

	# For debugging
	#Write-Host "Found $clips_cnt total items`n"

	# Initialize collectors
	$rtn_obj = [ordered]@{}

	# Walk through alerts list
	if ($LatestN) {
		$clips_cnt = $LatestN
	}
	
	# Initialize collector's keys
	foreach ($prop in $clipProps) {
		$rtn_obj.Add($prop, @())
	}
	
	for ($i=0; $i -lt $clips_cnt; $i++) { 
			
		if ($ClipPath) {
			if ($all_clips_paths[$i] -match $ClipPath) {
				$clip_path_idx = $i
				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host "Found match for ClipPath ($ClipPath)"
				}
			} else {
				Continue
			}
		}
			
		if ($FileType) {
			if ($all_clips_filetypes[$i] -ne $FileType) {
				Continue
			}
		}

		if ($clipProps.Count -gt 0) {
			
			# GET REQUESTED PROPERTIES
			
			foreach ($prop in $clipProps) {
				
				#Write-Host "clipProp = $prop"
				
				$arr = $clip_props_obj[$prop]

				# For debugging
				#Write-Host "arr = $arr"

				if ($arr) {
					
					$val = $arr[$i]

					# For debugging
					#Write-Host "val = $val"
					
					# Append value to collector
					if ($clipProps.Count -ge 1) {
						if ($val -or $val -eq 0) {
							$rtn_obj.$prop += $val				
						} else {
							$rtn_obj.$prop += $null
						}

						# For debugging
						#Write-Host ("rtn_obj[$prop] = " + ($rtn_obj.$prop | ConvertTo-Json))
						
					} else {
						if ($val -or $val -eq 0) {
							$rtn_obj = $val					
						} else {
							$rtn_obj = $null
						}
					}

					# For debugging
					#Write-Host ("item #" + ($i+1) + " $prop = $val")

				} else {
					if ($i -eq 0) {
						if ($FromConsoleFlg) {
							Write-Warning "$fn_name() nonexistent ClipProp '$prop' requested"
						}
					}
				}
			}
			
		} else {
			
			# GET ALL PROPERTIES

			# Get all property names
			$clip_props_obj_keys = $clip_props_obj.Keys

			# For debugging
			#Write-Host ("clip_props_obj_keys = " + $clip_props_obj_keys)
			
			foreach ($key in $clip_props_obj_keys) {

				# Initialize collector's keys
				if (($i -eq 0) -or ($i -eq $clip_path_idx)) {
					$rtn_obj.Add($key, @())
				}

				# Get all values for the property (an array)
				$vals = $clip_props_obj[$key]
				
				if ($vals) {
					$val = $vals[$i]
					$rtn_obj.$key = $rtn_obj.$key + $val					
				} else {
					$rtn_obj.$key = $rtn_obj.$key + $null
				}
				
				# For debugging
				#Write-Host ("item #" + ($i+1) + " $key = $val")
			}
		}
	}
	
	
	# @@@ DEV - Retain for testing
	#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)

	return $rtn_obj
}

function ResizeImage() {
# PURPOSE:  Resize image file to be used as attachment for Pushover notification 
# ARGS:     see param lines below
# RETURNS:  (array) @(before_KB, after_KB, before_WH, after_WH)
# NOTES:    Be sure to check if 'after_KB' = 0, as this indicates the new file was not actually created
# ATTRIBUTIONS:
#   Derived from: https://www.lewisroberts.com/2015/01/18/powershell-Grab-resize-function/
#   Also used: https://gist.github.com/someshinyobject/617bf00556bc43af87cd
	
    param(
		[string]$SourceFilespec,          # full filespec
		[string]$NewImagePath,            # folder location for resized files
		[int]$Quality = 90,               # integer 0 to 100
		[int]$Width = 500,                # pixels
		[int]$TargetKb,                   # kilobytes; NOTE: Pushover API max is 2,621,440 bytes (2621 KB)
		[string]$TargetKbMethod = 'WxH',  # 'WxH', 'Quality', or 'Hybrid'
		[int]$TargetKbWxHDelta = 25,      # percentage to decrease WxH per iteration
		[int]$TargetKbQualityDelta = 25,  # percentage to decrease Quality per iteration
		[boolean]$SilentFlg               # $true prevents background messaging (showing file resizing results)
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"	
	#Write-Host "$fn_name() SourceImagePath = '$SourceFilespec'"
	#Write-Host "$fn_name() NewImagePath = '$NewImagePath'"
	#Write-Host "$fn_name() Quality = $Quality"
	#Write-Host "$fn_name() Width = $Width"
	#Write-Host "$fn_name() TargetKb = $TargetKb"
	#Write-Host "$fn_name() TargetKbMethod = $TargetKbMethod"
	#Write-Host "$fn_name() TargetKbWxHDelta = $TargetKbWxHDelta"
	#Write-Host "$fn_name() TargetKbQualityDelta = $TargetKbQualityDelta"
	
	if (-not(Test-Path -LiteralPath "$SourceFilespec") ) {
		$oldSizeKb = $null
	}
	
	if ($Quality -eq 100 -and $Width -eq 0) {
		$resize_flg = $false
	} else {		
		$resize_flg = $true
	}
	#Write-Host "$fn_name() resize_flg = $resize_flg"

	$OutputLocation = $NewImagePath
	
    Add-Type -AssemblyName "System.Drawing"
    $img = [System.Drawing.Image]::FromFile($SourceFilespec)
 
	$oldHeight = $img.Height
	$oldWidth  = $img.Width
	$oldSizeKb = [math]::Round((Get-Item -Path $SourceFilespec).Length/1kB,1).ToString("#.0")
	$oldWH = "$oldWidth x $oldHeight"
		
	if ($script:fromConsole) {
		Write-Host "SourceImagePath = '$SourceFilespec'"
		Write-Host "orig height/width/size(KB) = $oldHeight/$oldWidth/$oldSizeKb"
	}
	
	# SPECIAL CASE - ITERATE RESIZING UNTIL FILE SIZE IS BELOW TARGET KILOBYTES
	if ($TargetKb -gt 0) {
		
		if ($script:fromConsole) {
			Write-Host "iterating to achieve file size < $TargetKb KB (using method '$TargetKbMethod')..."
		}
		
		if ($TargetKbMethod -eq 'WxH') {
			$i = 1
			$step = $TargetKbWxHDelta
			for ($x=(100-$step); $x -gt 0; $x=$x-$step) {
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host ("trying size reduction = -" + (100-$x) + "%")
				}
				
				$k = $x/100
				$CanvasHeight = $oldHeight*$k
				$CanvasWidth  = $oldWidth*$k

				# Encoder parameter for image quality
				$GrabEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($GrabEncoder, $Quality)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
			 
				# Compute the final ratio to use
				$ratioX = $CanvasWidth / $img.Width;
				$ratioY = $CanvasHeight / $img.Height;
			 
				$ratio = $ratioY  # Default
				if ($ratioX -le $ratioY) {
					$ratio = $ratioX
				}
			 
				$newWidth = [int] ($img.Width * $ratio)
				$newHeight = [int] ($img.Height * $ratio)

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($scriot:post_to_BI_logfile_flg) {
						PostCustomBiLogEntry -MsgStr "$err_msg ($script:ScriptName)" -Level 2 -SilentFlg $SilentFlg
					}
					if ($script:fromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")

				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [int]$TargetKb) {
					if ($script:fromConsole) {
						Write-Host "achieved target after $i iterations"
					}
					break				
				} else {
					if ($x -le $step) {
						if ($script:fromConsole) {
							Write-Host ("unable to achieve target after $i iterations; using final try results")
						}
					}
				}	
				$i++
			}
		}
		
		if ($TargetKbMethod -eq 'Quality') {
			$i = 1
			$step = $TargetKbQualityDelta  # percentage to decrease WxH per iteration
			for ($q=($Quality-$step); $q -gt 0; $q=$q-$step) {
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host ("trying quality reduction = -" + ($i*$step) + "%")
				}

				# Encoder parameter for image quality
				$GrabEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($GrabEncoder, $q)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
				$newWidth = $img.Width
				$newHeight = $img.Height

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($scriot:post_to_BI_logfile_flg) {
						PostCustomBiLogEntry -MsgStr "$err_msg ($script:ScriptName)" -Level 2 -SilentFlg $SilentFlg
					}
					if ($script:fromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")
					
				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [int]$TargetKb) {
					if ($script:fromConsole) {
						Write-Host "achieved target after $i iterations"
					}
					break				
				} else {
					if ($q -le $step) {
						if ($script:fromConsole) {
							Write-Host ("unable to achieve target after $i iterations; using final try results")
						}
					}
				}	
				$i++
			}
		}
			
		if ($TargetKbMethod -eq 'Hybrid') {
			$i = 1
			$step = $TargetKbWxHDelta
			for ($x=(100-$step); $x -gt 0; $x=$x-$step) {
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host ("trying hybrid reduction: size = -" + (100-$x) + "%; quality = -" + ($i*$TargetKbQualityDelta) + "%")
				}
				
				$j = $x/100
				$CanvasHeight = $oldHeight*$j
				$CanvasWidth  = $oldWidth*$j
				$q = $Quality - $i*$TargetKbQualityDelta

				# Encoder parameter for image quality
				$GrabEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($GrabEncoder, $q)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
			 
				# Compute the final ratio to use
				$ratioX = $CanvasWidth / $img.Width;
				$ratioY = $CanvasHeight / $img.Height;
			 
				$ratio = $ratioY  # Default
				if ($ratioX -le $ratioY) {
					$ratio = $ratioX
				}
			 
				$newWidth = [int] ($img.Width * $ratio)
				$newHeight = [int] ($img.Height * $ratio)

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($scriot:post_to_BI_logfile_flg) {
						PostCustomBiLogEntry -MsgStr "$err_msg ($script:ScriptName)" -Level 2 -SilentFlg $SilentFlg
					}
					if ($script:fromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")
					
				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:fromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [int]$TargetKb) {
					if ($script:fromConsole) {
						Write-Host "achieved target after $i iterations"
					}
					break				
				} else {
					if (($x -le $step) -or ($q -le $TargetKbQualityDelta)) {
						if ($script:fromConsole) {
							Write-Host ("unable to achieve target after $i iterations; using final try results")
						}
					}
					# Look ahead (necessary b/c $q is not managed by for:next loop)
					if (($q - $TargetKbQualityDelta) -lt $TargetKbQualityDelta) {
						break
					}
				}	
				$i++
			}
		}

		if ($resize_flg) {
			if ($script:fromConsole) {
				Write-Host "new height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				Write-Host "NewImagePath = '$NewImagePath'"
			}
		}
				
		# Release resource to avoid memory leak
		$img.Dispose()  
		
		return @($oldSizeKb, $newSizeKb.ToString("#.0"), $oldWH, $newWH)

	} 
	
	# ALL OTHER CASES
	if ($TargetKB -eq 0) {
		
		if (($Quality -eq 100) -and ($Width -eq 0)) {

			# USE THE SOURCE IMAGE AS-IS

			$newHeight = $oldHeight
			$newWidth  = $oldWidth
			$newSizeKb = $oldSizeKb
			
			# Copy the original image file 'as-is'
			Copy-Item -LiteralPath "$SourceFilespec" -Destination "$NewImagePath"			
			# ... and give it a current timestamp
			$file = Get-Item "$NewImagePath"
			$file.LastWriteTime = (Get-Date)	

		} else {

			# RESIZE THE SOURCE IMAGE
		 
			# Encoder parameter for image quality
			$GrabEncoder = [System.Drawing.Imaging.Encoder]::Quality
			$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
			$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($GrabEncoder, $Quality)
		 
			# Get codec
			$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
		 
			if (($Quality -lt 100) -and ($Width -eq 0)) {
				# User requested quality change w/o specifying a width change
				$CanvasWidth  = $oldWidth
				$CanvasHeight = $oldHeight
				$ratioX = 1
				$newWidth  = $oldWidth
				$newHeight = $oldHeight
			} else {
				$CanvasWidth  = $Width
				$CanvasHeight = ($Width / $oldWidth) * $oldHeight
				# Compute the final ratio to use
				$ratioX = $CanvasWidth / $img.Width
				$newWidth  = [int] ($img.Width * $ratioX)
				$newHeight = [int] ($img.Height * $ratioX)
			}
			
			# For debugging
			#Write-Host "CanvasWidth = $CanvasWidth"
			#Write-Host "CanvasHeight = $CanvasHeight"
			#Write-Host "ratioX = $ratioX"
			#Write-Host "newWidth = $newWidth"
			#Write-Host "newHeight = $newHeight"

			$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
			#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
			$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
			$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
		 
			$graph.Clear([System.Drawing.Color]::White)
			$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
		 
			# Save to file
			$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
			$bmpResized.Dispose()
			$img.Dispose()
		
			if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
				$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
				if ($script:post_to_BI_logfile_flg) {
					PostCustomBiLogEntry -MsgStr "$err_msg ($script:ScriptName)" -Level 2 -SilentFlg $SilentFlg
				}
				if ($script:fromConsole) {
					Write-Host $err_msg
				}
				msg.exe * /TIME:10 $err_msg
				
				# Release resource to avoid memory leak
				#$img.Dispose()  

				return @($oldSizeKb, $null)
			} else {
				$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1).ToString("#.0")
			}
		}
		if ($resize_flg) {
			if ($script:fromConsole) {
				Write-Host "NewImagePath = '$NewImagePath'"
				Write-Host "new height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
			}
		}
	}
	
	$oldWH = "$oldWidth x $oldHeight"
	$newWH = "$newWidth x $newHeight"
	
	# For debugging
	#Write-Host "oldWH = $oldWH"
	#Write-Host "newWH = $newWH"
	
	# Release resource to avoid memory leak
	#$img.Dispose()
	
	return @($oldSizeKb, $newSizeKb, $oldWH, $newWH)
}

function SendPushover() {
# PURPOSE: Send custom notification via the Pushover API
# ARGS:    $MsgTitle      = (string) pushover notification title
#          $MsgStr        = (string) pushover notification message; may contain HTML
#          $AttachPath    = (string) full filespec of resized alert image, or $null
#          $PriorityLevel = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency (default=0)
#          $SilentFlg     = (boolean) $true prevents background messaging (showing the message sent)
# RETURNS: nothing
# ATTRIBUTIONS:
#    handling attachments: https://tinyurl.com/mrxp7pn3  (PoshOver Powershell Package)
	
    param(
		[string]$MsgTitle,    # may contain HTML
		[string]$MsgStr,      # may contain HTML
		[string]$AttachPath,  # full filespec of resized alert image
		[int]$PriorityLevel,  # -2|-1|0|1|2
		[boolean]$SilentFlg
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
	#Write-Host "$fn_name() MsgStr = '$MsgStr'"
	#Write-Host "$fn_name() AttachPath = '$AttachPath'"	
	#Write-Host "$fn_name() PriorityLevel = PriorityLevel"
	#Write-Host "$fn_name() SilentFlg = $SilentFlg"
	#Write-Host "$fn_name() fromConsole = $script:fromConsole"
	# NOTE REGARDGING MSG.EXE COMMANDS
	# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
	#msg.exe * /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nSilentFlg = $SilentFlg`nAttachPath = $AttachPath`nMsgStr = $MsgStr"
	
	#Prep the attachment
	if ($AttachPath) {
		#Get-Item -Path $AttachPath
		$attach_obj = (Get-Item -Path $AttachPath)
		$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
	}

	# For debugging
	#Write-Host ("attach_obj = " + $attach_obj)
	#Write-Host ("attach_obj.Name = " + $attach_obj.Name)
	
	$uri = 'https://api.pushover.net/1/messages.json'
	$parameters = @{
		token      = $script:po_app_token
		user       = $script:po_user_token
		priority   = $PriorityLevel
		retry      = $script:po_retry
		expire     = $script:po_expire
		tags       = $script:po_tags
		sound      = $script:po_sound
		device     = $script:po_device
		html       = 1
		title      = $MsgTitle
		message    = $MsgStr
	}
	
	# For debugging
	#Write-Host ("parameters.Keys = " + $parameters.Keys)
	#Write-Host "`nPushover parameters..."
	[string[]]$arr = $parameters | Out-String -stream
	foreach ($item in $arr) {
		if ($item) {
			#Write-Host $item
		}
	}
	
	# EXECUTE
	if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
		# SEND BASIC NOTIFICATION

		# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
		#	  do not execute; it doesn't produce anything, no error, warning, or output.
		#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
		#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
		#     all web requests now use basic parsing only.
		# ... attachmnets
		#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
		$Response = $parameters | Invoke-RestMethod -Uri $uri -Method Post
		# Alternative syntaxes
		#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
		#$Response = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters

	} else {	
		# SEND NOTIFICATION WITH ATTACHMENT
		
		# Notes:
		# Borrowed code from:  https://github.com/joshooaj/Poshover/blob/main/src/Private/Send-MessageWithAttachment.ps1
		# More refs:           https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

		try {

            Add-Type -AssemblyName System.Net.Http
			# NOTE: Originally, the $client= line below was not working when this script was 
			#        executed in the background via a Blue Iris 'run a program' action
			#        SOLUTION:  https://stackoverflow.com/a/70535083
			$client = [system.net.http.httpclient]::new()

			try {
				$content = [system.net.http.multipartformdatacontent]::new()
				foreach ($key in $parameters.Keys) {
					$textContent = [system.net.http.stringcontent]::new($parameters.$key)
					$content.Add($textContent, $key)
				}
				
				# For debugging
				#Write-Host ("content = " + $content | ConvertTo-Json)
				
				$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
				$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
				$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
				$jpegContent.Headers.ContentDisposition.Name = 'attachment'
				$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
				$content.Add($jpegContent)

				# Not needed.
				# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
				#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

				# Modify the message body
				Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

				# DO IT!
				#msg.exe * /TIME:10 "executing pushover api cmd w/ attachment..."
				$Response = $client.PostAsync($uri, $content).Result
				
				if ($script:fromConsole) {
					Write-Output ($Response.Content.ReadAsStringAsync().Result | ConvertFrom-Json)
				}
			}
			finally {
				$content.Dispose()
			}
		}
		finally {
			$client.Dispose()
		}
	}

	# PROVIDE FEEDBACK
	if (!$SilentFlg) {
		#msg.exe * /TIME:15 "Pushover notification sent...`npriority_level = $PriorityLevel`n$MsgStr"
		#msg.exe * /TIME:10 "Pushover notification attachment...`n$AttachPath"
		#msg.exe * /TIME:30 ("Pushover notification response...`n" + ($Response | ConvertTo-Json))
		#msg.exe * /TIME:30 ("Pushover notification response...`n" + $Response)
	}
	
	# For debugging
	#if (!((($Response.Content.ReadAsStringAsync().Result | ConvertFrom-Json)) -match "invalid")) {
		
	if ((($Response.Content.ReadAsStringAsync().Result | ConvertFrom-Json)) -match "invalid") {
		$err_msg = "ERR/Pushover API returned 'invalid'"
		if ($script:post_to_BI_logfile_flg) {
			PostCustomBiLogEntry -MsgStr "$err_msg ($script:ScriptName)" -Level 2 -SilentFlg $SilentFlg
		}
		if ($script:fromConsole) {
			Write-Host "$err_msg`n"
		}
		if (!($SilentFlg)) {
			msg.exe * /TIME:10 "$err_msg ($script:ScriptName)"			
		}
	} else {
		if ($script:fromConsole) {
			Write-Host "Pushover notification sent`n"
		}
	}
}

function LogPerformance() {
# PURPOSE:  Log performance data to CSV file for image file resizing and Pushover notification actions.
# ARGS:     see param lines below
# RETURNS:  nothing
# NOTES:    1. CSV file is created automatically if it does not already exist. 
#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
	
    param(
		[string]$csv_filespec, 
		[string]$img_filename, 
		[string]$timestamp, 
		[double]$before_kb, 
		[double]$after_kb, 
		[string]$before_wh, 
		[string]$after_wh, 
		[double]$total_ms, 
		[double]$resize_ms, 
		[double]$pusho_ms
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
		
	# Build the new row (object keys are the header column labels)
	$new_row_obj = [PSCustomObject]@{
		'Timestamp'    = $timestamp
		'SizeBeforeKb' = $before_kb
		'SizeAfterKb'  = $after_kb
		'WxHBefore'    = $before_wh
		'WxHAfter'     = $after_wh
		'TotalMsec'    = $total_ms
		'ResizeMsec'   = $resize_ms
		'PushoMsec'    = $pusho_ms
		'ImageFile'    = $img_filename
	}
	
	# For debugging
	#msg.exe * /TIME:10 "new_row_obj... `n$new_row_obj"

	# Append new row to CSV file, creating the file if necessary
	if ($PSVersionTable.PSVersion.Major -ge 7) {
		# REFS: https://tinyurl.com/yvy4mjy6
		#       https://stackoverflow.com/a/70717331
		#       https://stackoverflow.com/a/60681762
		$new_row_obj | Export-Csv -LiteralPath $csvfilespec -Append -Force -UseQuotes AsNeeded
	} else {
		$new_row_obj | Export-Csv -LiteralPath $csvfilespec -Append -Force -NoTypeInformation
	}
}

# Start overall performance timer
$timer_total = [Diagnostics.Stopwatch]::StartNew()

if ($Help) {
	if ($fromConsole) {
		Write-Host "==========================================================================================="
		Write-Host "Send Pushover Notification with Current Camera or Group Image"
		Write-Host "By default, script takes, then sends Blue Iris snapshot (camera only)"
		Write-Host "Alternatively, script can grab, then send current image from video stream (camera OR GROUP)"
		Write-Host "-------------------------------------------------------------------------------------------"
		Write-Host "Usage: $ScriptName -Msg 'current image' -CamName 'FrontDoor'"
		Write-Host "Arguments:"
		Write-Host "   -Msg        Pushover notification message (may contain HTML)"
		Write-Host "   -CamName    camera shortname (or camera group name, if using -Grab)"
		Write-Host "   -Title      customize Pushover notification title (default is 'from $ScriptName)"
		Write-Host "   -Priority   Pushover notification priority: -2|-1|0|1|2 = lowest|silent|normal|high|emergency"
		Write-Host "   -Grab       grab substream image iNSTEAD OF TAKING SNAPSHOT (enables sending cam group image)"
		Write-Host "   -Mainstream w/ -Grab, grabs mainstream image (cam group images are always substream)"
		Write-Host "   -Quality    customize quality of Pushover attachment (0 to 100)"
		Write-Host "   -Width      customize width of Pushover attachment"
		Write-Host "   -TargetKB   customize size of Pushover attachment (no larger than target)"
		Write-Host "   -Demo       do not send notification or log event to Blue Iris logfile (useful when testing)"
		Write-Host "   -NoBiLog    do not log event to Blue Iris logfile"
		#Write-Host "Note:  using '-Quality 100 -Width 0' sends snapshot without resizing"
		Write-Host "-------------------------------------------------------------------------------------------"
	}
}

# HANDLE COMMANDLINE ARGS
if ($Msg -and $CamName) {
	$po_msg = $Msg
} else {
	$err_msg = "ERR/expecting 2 arguments (-Msg -CamName)"
	if ($post_to_BI_logfile_flg) {
		PostCustomBiLogEntry -MsgStr "$err_msg ($ScriptName)" -Level 2 -SilentFlg ($bg_msgs_flg -eq $false)
	}
	if ($fromConsole) {
		return "$err_msg`n"
	} else {
		msg.exe * /TIME:10 $err_msg
	}	
}
if ($Title) {
	$po_title = $Title
}
if ($Mainstream -and !$Grab) {
	if ($fromConsole) {
		Write-Warning "argument -Mainstream ignored (intended for use with -Grab argument)"
	}
}
if ($Priority -or ($Priority -eq 0)) {
	if ($Priority -ne 0) {
		if (($Priority -ge -2) -and ($Priority -le 2)) {
			$po_priority = $Priority
		}
	}
}
#Write-Host "attach_path = '$attachment_path'"
$resize_flg = ($Quality -or $Width -or $TargetKB) 
if (!$resize_flg) {
	$img_quality = 100
	$img_width = 0
	$img_kilobytes = 0
}
#Write-Host "resize_flg = '$resize_flg'"
if ($Quality) {
	if (([int]$Quality -ge 1) -and ([int]$Quality -le 100)) {
		$img_quality = [int]$Quality
	}
}
if ($Width -or ([int]$Width -eq 0)) {
	$img_width = [int]$Width
}
if ($TargetKB -or ([int]$TargetKB -eq 0)) {
	$img_kilobytes = [int]$TargetKB
}
if ($Quiet) {
	$bg_msgs_flg = $false
}

# MANAGE SCRIPT'S FOLDERS
if (!(Test-Path -LiteralPath "$app_path")) {
	New-Item -ItemType Directory -Force -Path $app_path | Out-Null
}

# GET PUSHOVER ATTACHMENT IMAGE IMAGE

$snap_file_found_flg = $false  # Assume failure
$attachment_path = "$app_path$CamName.jpg"

if ($Grab) {

	# GRAB IMAGE FROM CAMERA STREAM
	$response = `
		GrabStreamImage `
		-CamName $CamName `
		-MainStreamFlg $Mainstream `
		-DestPath $attachment_path

	# For debugging
	#Write-Host "GrabStreamImage() response = '$response'"
	
	if ($response -eq 'success') {
		$snap_file_found_flg = $true
	}
	
} else {
	
	# TAKE CAMERA SNAPSHOT
	$response = `
		TakeSnapshot `
		-CamName $CamName

	# For debugging
	#Write-Host "TakeSnapshot() response = $response"
	#msg.exe $script:server_username /TIME:60 "$fn_name() trying to get most recent DAT file"

	# Necessary to avoid a race condition where the script looks for the latest snapshot JPG file 
	# before the Windows filesystem is ready
	Start-Sleep -Milliseconds $snap_delay_ms

	# GET SNAPSHOT FILE
	if ($response -eq 'success') {
				
		# METHOD 1 - Search New folder for camera's most recent JPG file 

		# Start performance timer
		$timer_method_1 = [Diagnostics.Stopwatch]::StartNew()

		$pattern = $bi_new_path + "$camName*.jpg"
		if ($PsVersion -ge 7) {
			if ($DevMode) {
				if ($fromConsole) {
					Write-Host "`n(dev mode) most recent JPG files..."
					(Get-ChildItem $pattern | Sort -Descending) | Select -First 3
				}
			}
			$latest_snap_jpg_filespec = (Get-ChildItem $pattern | Sort -Descending) | Select -First 1
		} else {
			if ($DevMode) {
				if ($fromConsole) {
					Write-Host "`n(dev mode) most recent JPG files..."
					((Get-ChildItem $pattern).FullName | Sort -Descending) | Select -First 3
				}
			}
			$latest_snap_jpg_filespec = ((Get-ChildItem $pattern).FullName | Sort -Descending) | Select -First 1
		}
		
		# For debugging
		#Write-Host "pattern = $pattern"
		#Write-Host "latest_snap_jpg_filespec = $latest_snap_jpg_filespec"

		# Make sure the snapshot JPG is really current (no older than a few seconds)
		if ($latest_snap_jpg_filespec) {
			$max_age_sec = 5
			$file_dt = (Get-Item -LiteralPath "$latest_snap_jpg_filespec").LastWriteTime
			$diff = (Get-Date).Subtract($file_dt)
			$age_sec = [math]::Floor($diff.TotalSeconds)

			if ($DevMode) {
				if ($fromConsole) {
					Write-Host "(dev mode) snapshot file is $age_sec seconds old ('$latest_snap_jpg_filespec')"
				}
			}
			
			if ($age_sec -lt $max_age_sec) {
				$snap_file_found_flg = $true
			}
		}

		# Stop performance timer
		$timer_method_1.stop()
		$method_1_ms = [math]::Round($timer_method_1.elapsed.totalmilliseconds,1).ToString("#.0")

		# For debugging
		#msg.exe $script:server_username /TIME:60 "$fn_name() dat_file2_found_flg = $dat_file_found_flg`ndat_file = '$dat_file'`nage_sec = $age_sec"
		
		# METHOD 2 - Query the Blue Iris database for the 'cliplist'

		# Start method 2 performance timer
		$timer_method_2 = [Diagnostics.Stopwatch]::StartNew()

		$clip_list_obj = `
			BIJsonCliplist `
			-CamName $CamName `
			-Props 'file,path,flags' `
			-FileType 'jpg  New' `
			-FromConsoleFlg $fromConsole `
			-QuietFlg $Quiet

		# For debugging
		#Write-Host ("clip_list_obj = " + ($clip_list_obj | ConvertTo-Json))
		
		$latest_cliplist_fname = $clip_list_obj['file'][0]
		$source_path = "$bi_new_path" + $latest_cliplist_fname

		# For debugging
		#Write-Host "snap source_path = '$source_path'
		
		# Save copy of the snapshot file to the script's working folder
		if ($source_path -eq $latest_snap_jpg_filespec) {
			if (Test-Path -LiteralPath "$source_path") {
				$snap_file_found_flg = $true
				Copy-Item -LiteralPath $source_path -Destination $attachment_path
			}
		}

		# Stop performance timer
		$timer_method_2.stop()
		$method_2_ms  = [math]::Round($timer_method_2.elapsed.totalmilliseconds,1).ToString("#.0")
		
		if ($DevMode) {
			if ($fromConsole) {
				Write-Host "(dev mode) snapshot get file method_1_ms = $method_1_ms"
				Write-Host "(dev mode) snapshot get file method_2_ms = $method_2_ms"
			}
		}

		# ADD CUSTOM MEMO TO DATABASE
		if ($snap_file_found_flg) {
			
			$jsonCmd = @{
				cmd = 'update'
				path = $clip_list_obj['path'][0]
				memo = $snap_memo
				flags = $clip_list_obj['flags'][0]
			}								
			# Returns a powershell PSCustomObject
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd `
				-QuietFlg $Quiet `
				-FromConsoleFlg $FromConsole

			# For debugging
			#Write-Host ("jsonCmd = " + ($jsonCmd | ConvertTo-Json))
			#Write-Host ("json_response_obj = " + ($json_response_obj | ConvertTo-Json))
		}
	}
}

# For debugging
#Write-Host "attach_path = '$attachment_path'"
#Write-Host "snap_file_found_flg = '$snap_file_found_flg'"

if ($snap_file_found_flg) {	

	#if ($response -eq 'success') {
		
		if (-not(Test-Path -LiteralPath "$attachment_path") ) {
			$err_msg = "ERR/snapshot image file not found! (expecting the full filespec)`n$attachment_path"
			if ($post_to_BI_logfile_flg) {
				PostCustomBiLogEntry -MsgStr "$err_msg ($ScriptName)" -Level 2 -SilentFlg ($bg_msgs_flg -eq $false)
			}
			if ($fromConsole) {
				return $err_msg
			} else {
				msg.exe * /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
			}	
		}

		# Create resized image's full filespec
		$full_filespec = (Resolve-Path $attachment_path).Path

		#Write-Host "full_filespec = $full_filespec"
		
		$dot_idx = $full_filespec.LastIndexOf(".")
		$resized_image_path = $full_filespec.Substring(0,$dot_idx) + "_resized" + $full_filespec.Substring($dot_idx,$full_filespec.Length - $dot_idx)

		# Change destination path, if requested
		$basename = Split-Path $attachment_path -leaf
		$source_path = $full_filespec.Replace($basename, "")
		$resized_image_path = $resized_image_path.Replace($source_path, $app_path)

		# For debugging
		#Write-Host "source image full_filespec = '$full_filespec'"
		#Write-Host "source image filename = '$basename'"
		#Write-Host "source image source_path = '$source_path'"
		#Write-Host "resized_image_path = '$resized_image_path'"
		#Write-Host "img_quality = '$img_quality'"
		#Write-Host "img_width = '$img_width'"
		#Write-Host "img_kilobytes = '$img_kilobytes'"

		# Start resize performance timer
		$timer_resize = [Diagnostics.Stopwatch]::StartNew()

		# RESIZE ATTACHMENT
		$resize_ary = `
			ResizeImage `
			-SourceFilespec $attachment_path `
			-NewImagePath $resized_image_path `
			-Quality $img_quality `
			-Width $img_width `
			-TargetKb $img_kilobytes `
			-TargetKbMethod $targetkb_method `
			-TargetKbWxHDelta $targetkb_wxh_step `
			-TargetKbQualityDelta $targetkb_quality_step `
			-SilentFlg ($bg_msgs_flg -eq $false)

		# Stop resize performance timer
		$timer_resize.stop()
		$resize_ms = [math]::Round($timer_resize.elapsed.totalmilliseconds,1).ToString("#.0")
	#}

	# Start pushover performance timer
	$timer_pusho = [Diagnostics.Stopwatch]::StartNew()

	# SEND NOTIFICATION
	if (!$Demo) {
		# Add info to notification message
		if ($response -eq 'success') {
			if ($show_method_flg) {
				if ($Grab) {
					if ($Mainstream) {
						$po_msg = $po_msg + "`n(method: mainstream grab)"
					} else {
						$po_msg = $po_msg + "`n(method: substream grab)"
					}
				} else {
					$po_msg = $po_msg + "`n(method: snapshot)"
				}
			}
			# Make another info string to be shown in dev mode
			if ($DevMode) {
				$arr = @()  # Initialize accumulator
				if ($Quality) {
					if (!($arr -contains "`n(")) {
						$arr += "`n("
					}
					$arr += "Quality=$img_quality"
				}
				if ($Width) {
					if (!($arr -contains "`n(")) {
						$arr += "`n("
					}
					if (($arr -Join "") -match "Quality") {
						$arr += "; "
					}
					$arr += "Width=$img_width"
				}
				if ($TargetKB) {
					if (!($arr -contains "`n(")) {
						$arr += "`n("
					}
					if ((($arr -Join "") -match "Quality") -or (($arr -Join "") -match "Width")) {
						$arr += "; "
					}
					$arr += "TargetKB=$img_kilobytes"
				}
				if ($arr.Count -gt 0) {
					$arr += ")"
				}
				# Sppend substring like... "`n(quality=90; width=500; targetKB=20)"
				$po_msg = $po_msg + ($arr -Join '')
			}
			if ($resize_ary[1] -eq 0) {
				# Try sending the original, unsized attachment
				SendPushover `
					-MsgTitle $po_title `
					-MsgStr $po_msg `
					-AttachPath $attachment_path `
					-PriorityLevel $po_priority `
					-SilentFlg ($bg_msgs_flg -eq $false)
					
				if ($fromConsole) {
					Write-Host "HEADS-UP: Requested attachment not resized; tried sending notification with uniresized file"
				}
			} else {
				# Send resized attachment
				SendPushover `
					-MsgTitle $po_title `
					-MsgStr $po_msg `
					-AttachPath $resized_image_path `
					-PriorityLevel $po_priority `
					-SilentFlg ($bg_msgs_flg -eq $false)
			}
		} else {
			# Send notification WITHOUT attachment
			SendPushover `
				-MsgTitle $po_title `
				-MsgStr $po_msg `
				-AttachPath $null `
				-PriorityLevel $po_priority `
				-SilentFlg ($bg_msgs_flg -eq $false)
		}
	}

	# Stop pushover performance timer
	$timer_pusho.stop()
	$pusho_ms  = [math]::Round($timer_pusho.elapsed.totalmilliseconds,1).ToString("#.0")

	# Stop overall performance timer
	$timer_total.stop()
	$total_ms  = [math]::Round($timer_total.elapsed.totalmilliseconds,1).ToString("#.0")

	# LOG SCRIPT PERFORMANCE INFO TO CSV FILE
	$csvfilespec = $app_path + $ScriptName + ".csv"
	$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
	$total_ms    = [decimal]$total_ms 
	$pusho_ms    = [decimal]$pusho_ms 
	if ($attachment_path) {
		$size_before = [decimal]$resize_ary[0]
		$size_after  = [decimal]$resize_ary[1]
		$wxh_before  = [string]$resize_ary[2]
		$wxh_after   = [string]$resize_ary[3]
		$resize_ms   = [decimal]$resize_ms 
		LogPerformance $csvfilespec $attachment_path $timestamp $size_before $size_after $wxh_before $wxh_after $total_ms $resize_ms $pusho_ms
	} else {
		$size_before = 0
		$size_after  = 0
		$wxh_before  = ""
		$wxh_after   = ""
		$resize_ms   = 0 
		LogPerformance $csvfilespec 'NA' $timestamp $size_before $size_after $wxh_before $wxh_after $total_ms $resize_ms $pusho_ms
	}

	# Show performance notice (dev mode)
	if ($fromConsole) {
		if ($DevMode) {
			$perf_msg = "(dev mode) script execution time = " + [math]::Round($timer_total.elapsed.totalmilliseconds,1) + " msec"
			if ($Demo) {
				Write-Host "`n$perf_msg"
			} else {
				Write-Host $perf_msg
			}
		}
	}

	# Format vales for messaging
	$wxh_before     = $wxh_before.Replace(" x ", "x")
	$wxh_after      = $wxh_after.Replace(" x ", "x")
	$disp_total_ms  = [math]::Round($total_ms,0)
	$disp_resize_ms = [math]::Round($resize_ms,0)
	$disp_pusho_ms  = [math]::Round($pusho_ms,0)

	# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
	if (!$fromConsole) {
		if ($bg_msgs_flg) {
			$msg_sec = 30
			$bg_msg =  $attachment_path + 
				"`n`nWxH before/after  = $wxh_before / $wxh_after" +
				"`n`nSize before/after (KB) ... $size_before / $size_after" + 
				"`n`nexecution times (resize,pusho,total) ... $disp_resize_ms, $disp_pusho_ms, $disp_total_ms msec" + 
				"`n`n(this msg auto-closes in $msg_sec seconds)"
			msg.exe * /TIME:$msg_sec $bg_msg
		}
	}

	# POST MESSAGE TO BLUE IRIS LOGFILE
	if (!$Demo -and !$NoBiLog) {
		if ($post_all_events) {
			$log_msg = "pushover snapshot [$wxh_before, $wxh_after; $size_before KB, $size_after KB; $disp_resize_ms msec, $disp_pusho_ms msec, $disp_total_ms msec] file=$attachment_path"
			PostCustomBiLogEntry -MsgStr "$log_msg" -Level 0 -SilentFlg $true
		}
	}
		
} else {
	if ($fromConsole) {
		Write-Warning "No notification sent - Unable to find current snapshot or image grab file"
	}
}

if ($fromConsole) {
	Write-Host ""
}

#Stop-Transcript